'''
Image Bounding Box Slicer
~~~~~~~~~~~~
Chop up images and its bounding boxes. 
Basic usage instructions are available in demo.ipynb

:copyright: (c) 2019 Akshay Chandra Lagandula.
:license: MIT, see LICENSE for more details.
'''

__title__ = 'image_bbox_slicer'
__author__ = 'Akshay Chandra Lagandula'
__license__ = 'MIT'
__copyright__ = 'Copyright 2019 Akshay Chandra Lagandula'

from .main import *
from .helpers import *
